<?php
/**
 * Plugin Name:         RS Redirect 404
 * Description:         Try to avoid unnecessary 404 errors
 * Plugin URI:          https://plugins.seindal.dk/plugins/rs-redirect-404/
 * Update URI:          https://plugins.seindal.dk
 * Author:              René Seindal
 * Author URI:          https://plugins.seindal.dk/
 * Donate link:         https://mypos.com/@historywalks
 * Requires Plugins:    rs-base-plugin
 * License:             GPL v2 or later
 * License URI:         https://www.gnu.org/licenses/gpl-2.0.html
 * Text Domain:         rs-redirect-404
 * Domain Path:         /languages
 * Requires PHP:        7.4
 * Requires at least:   5.0
 * Version:             0.9.2
 **/

namespace ReneSeindal;

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

add_action( 'plugins_loaded', function() {
    class RS_Redirect404 extends PluginBase {
        protected $plugin_file = __FILE__;

        use PluginBaseSettings;

        protected $default_options = [
            'maxdist' => 3,
            'blacklist' => '',
        ];

        // Don't bother with these URIs
        protected $fixed_blacklist = [
            '/.well-known/',
            '/nodeinfo/',
            '/api/',
            '/wp-json/',
        ];


        // Redirect
        function redirect( $target, $uri ) {
            if ( empty( $target ) ||  $target == site_url( $uri ) )
                return;

            wp_redirect( $target, 301 );
            exit;
        }

        // Redirect url to a post
        function redirect_target_post( $post ) {
            return get_permalink( $post );
        }

        // Redirect url to a image
        function redirect_target_image( $image, $size = 'full' ) {
            return wp_get_attachment_image_url( $image, $size );
        }

        // Redirect url to a term listing
        function redirect_target_term( $term ) {
            return get_term_link( $term );
        }


        function do_template_redirect_action() {
            if ( !is_404() ) return;

            $uri = NULL;
            if ( isset( $_SERVER['REQUEST_URI'] ) && $_SERVER['REQUEST_URI'] )
                $uri = sanitize_url( urldecode( $_SERVER['REQUEST_URI'] ) );

            if ( $uri ) {
                foreach ( $this->fixed_blacklist as $skip )
                    if ( str_starts_with( $uri, $skip ) )
                        return;

                $blacklist = $this->get_option( 'blacklist' );
                if ( $blacklist ) {
                    foreach ( explode( "\n", $blacklist ) as $skip )
                        if ( str_starts_with( $uri, $skip ) )
                            return;
                }
            }

            global $wp_query;

            // If WP thinks the request is for a post of some type we have 'name'

            $name = $wp_query->get( 'name' );
            if ( $name ) {
                $target = $this->try_image_search( $uri );
                if ( !$target )
                    $target = $this->try_post_search( $name );

                return $this->redirect( $target, $uri );
            }


            // If WP thinks the request is for a taxonomy term

            $tax_query = $wp_query->tax_query;

            if ( $tax_query ) {
                $queried = $tax_query->queried_terms;

                foreach ( $queried as $taxonomy => $data ) {
                    if ( is_array( $data['terms'] ) ) {
                        foreach ( $data['terms'] as $slug ) {
                            $target = $this->try_term_search( $taxonomy, $slug );
                            if ( $target )
                                return $this->redirect( $target, $uri );
                        }
                    }
                }
            }
        }

        function try_image_search( $uri ) {
            // Is this a resized image request?
            if ( preg_match( '/^(.*)-(\d+)x(\d+)(\.\w+)$/' , $uri, $m ) ) {

                // There are the possible paths that I'm aware of
                $paths = [
                    $m[1] . $m[4],
                    $m[1] . '-rotated' . $m[4],
                    $m[1] . '-scaled' . $m[4],
                ];
                $width = $m[2];
                $height = $m[3];

                foreach ( $paths as $path ) {
                    $image_uri = site_url( $path );
                    $image_id = attachment_url_to_postid( $image_uri );
                    if ( $image_id )
                        break;
                }

                if ( $image_id ) {
                    foreach ( wp_get_registered_image_subsizes() as $size => $data ) {
                        if ( ( ( empty( $data['width'] ) || $data['width'] >= $width ) )
                             &&
                             ( ( empty( $data['height'] ) || $data['height'] >= $height ) ) )
                            return $this->redirect_target_image( $image_id, $size );
                    }

                    return $this->redirect_target_image( $image_id );
                }
            }

            return NULL;
        }

        function try_post_search( $name ) {
            // Any post matching $name
            $posts = get_posts( [
                'post_status' => 'publish',
                'post_type' => get_post_types( [ 'publicly_queryable' => true ] ),
                'name' => $name,
            ] );

            // Look for past slugs
            if ( empty( $posts ) ) {
                $posts = get_posts( [
                    'post_status' => 'publish',
                    'post_type' => get_post_types( [ 'publicly_queryable' => true ] ),
                    'meta_key' => '_rs404_past_slug',
                    'meta_value' => $name,
                ] );
            }

            // If we've found something, redirect to the newest
            if ( $posts )
                return $this->redirect_target_post( $posts[0] );



            // See if we have a taxomony term matching - just in case
            $terms = get_terms( [
                'taxonomy' => get_taxonomies( [ 'public' => true ] ),
                'slug' => $name,
                'fields' => 'ids',
            ] );

            // If we've found something, redirect to the first
            if ( $terms )
                $this->redirect_target_term( $terms[0] );



            $maxdist = $this->get_option( 'maxdist' );
            if ( !$maxdist ) return;

            // Levenshtein search - expensive

            $ids = get_posts( [
                'post_status' => 'publish',
                'post_type' => get_post_types( [ 'publicly_queryable' => true ] ),
                'numberposts' => -1,
            ] );

            $best_score = strlen( $name );
            $best_post = NULL;

            foreach ( $ids as $post_id ) {
                $post = get_post( $post_id );
                $score = levenshtein( $name, $post->post_name );
                if ( $score < $best_score ) {
                    $best_score = $score;
                    $best_post = $post;
                }
            }

            if ( $best_score <= $maxdist )
                return $this->redirect_target_post( $best_post );

            return NULL;
        }

        function try_term_search( $taxonomy, $slug ) {
            // Search for term with past slug

            $terms = get_terms( [
                'taxonomy' => $taxonomy,
                'meta_key' => '_rs404_past_slug',
                'meta_value' => $slug,
                'fields' => 'ids',
            ] );

            if ( $terms )
                return $this->redirect_target_term( $terms[0] );


            // Levenshtein search - expensive

            $maxdist = $this->get_option( 'maxdist' );
            if ( !$maxdist ) return;

            $slugs = get_terms( [
                'taxonomy' => $taxonomy,
                'fields' => 'slugs',
            ] );

            $best_score = $maxdist + 1;
            $best_slug = NULL;

            foreach ( $slugs as $termslug ) {
                $score = levenshtein( $termslug, $slug );
                if ( $score < $best_score ) {
                    $best_score = $score;
                    $best_slug = $termslug;
                }
            }

            if ( $best_score <= $maxdist ) {
                $term = get_term_by( 'slug', $best_slug, $taxonomy );
                return $this->redirect_target_term( $term );
            }

            return NULL;
        }

        /************************************************************************
         *
         * Keep track of past slugs on all published posts - using post meta
         *
         * _rs404_cur_slug (unique) - is the current slug
         * _rs404_past_slug (array) - past slugs
         *
         ************************************************************************/

        // This is called before an existing post is updated

        function do_pre_post_update_action( $post_id, $data ) {
            if ( ! is_post_publicly_viewable( $post_id ) )
                return;

            $post = get_post( $post_id );
            $slug = $post->post_name;

            if ( $slug ) {
                update_post_meta( $post_id, '_rs404_cur_slug', $slug );
            }
        }

        // This is called after a post is saved - updated or created

        function do_save_post_action( $post_id, $post, $update ) {
            if ( ! is_post_publicly_viewable( $post_id ) )
                return;

            $post = get_post( $post_id );
            $slug = $post->post_name;

            if ( $slug ) {
                $prev = get_post_meta( $post_id, '_rs404_cur_slug', true );

                if ( $prev && $prev != $slug ) {
                    // Save the past slugs
                    $past = get_post_meta( $post_id, '_rs404_past_slug' );
                    if ( ! in_array( $prev, $past ?: [] ) )
                        add_post_meta( $post_id, '_rs404_past_slug', $prev );
                }

                // Set new current slug to check for future changes
                update_post_meta( $post_id, '_rs404_cur_slug', $slug );
            }
        }

        /************************************************************************
         *
         * Keep track of past slugs on all taxonomy terms - using post meta
         *
         * _rs404_cur_slug (unique) - is the current slug
         * _rs404_past_slug (array) - past slugs
         *
         ************************************************************************/

        // This is called before an existing term is updated

        function do_wp_update_term_data_filter( $data, $term_id, $taxonomy, $args ) {
            if ( ! is_term_publicly_viewable( $term_id ) )
                return;

            $term = get_term( $term_id );
            $slug = $term->slug;
            update_term_meta( $term_id, '_rs404_cur_slug', $slug );
            return $data;
        }

        // This is called after a term is save - updated or created

        function do_saved_term_action( $term_id, $tt_id, $taxonomy, $args ) {
            if ( ! is_term_publicly_viewable( $term_id ) )
                return;

            $term = get_term( $term_id );

            $slug = $term->slug;
            $prev = get_term_meta( $term_id, '_rs404_cur_slug', true );

            if ( $prev && $prev != $slug ) {
                // Save the past slugs
                $past = get_term_meta( $term_id, '_rs404_past_slug' );
                if ( ! in_array( $prev, $past ?: [] ) )
                    add_term_meta( $term_id, '_rs404_past_slug', $prev );
            }

            // Set new current slug to check for future changes
            update_term_meta( $term_id, '_rs404_cur_slug', $slug );
        }

        /************************************************************************
         *
         *	Settings
         *
         ************************************************************************/

        function do_admin_menu_action() {
            $this->settings_add_menu(
                __( 'RS Redirect404', 'rs-redirect-404' ),
                __( 'RS Redirect404', 'rs-redirect-404' )
            );
        }

        function settings_define_sections_and_fields() {

            $section = $this->settings_add_section(
                'section_levenshtein',
                __( 'Levenshtein (fuzzy) searching', 'rs-redirect-404' ),
                __( 'Fuzzy searching for post and term slugs - can be CPU intensive.', 'rs-redirect-404' )
            );


            $this->settings_add_field(
                'maxdist', $section,
                __( 'Maximum Levenshtein distance', 'rs-redirect-404' ),
                'settings_field_select_html',
                [
                    'values' => [
                        0 => 'Don\' use',
                        1 => '1',
                        2 => '2',
                        3 => '3',
                        4 => '4',
                        5 => '5',
                    ],
                ]
            );


            $section = $this->settings_add_section(
                'section_blacklist',
                __( 'Blacklist settings', 'rs-redirect-404' ),
                __( 'List of URLs to skip searching for - must match exactly at the start of the URL', 'rs-redirect-404' )
            );

            $this->settings_add_field(
                'blacklist', $section,
                __( 'URL blacklist', 'rs-redirect-404' ),
                'settings_field_textarea_html'
            );
        }

    }

    RS_Redirect404::instance();
} );
